/* Put in the public domain 2000 by Sam Trenholme */
/* Kiwi Main:  The part of the Kiwi rouines with main() in it and the
   relevent command line parsers */

#include "JessStr/JessStr.h"
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <time.h>

/* Function prototypes */
void gencookie();
void decode();
void wrapper();
void infilter();

/* Alphanumberic digits */
#define AN "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz1234567890-_"

/* Size of cookie in octets */
#define COOKIE_SIZE 7

/* Softerror: Exit with an error code allowing the MTA to try again */
int softerror(char *errormsg) {
    fprintf(stderr,"Soft error: %s\n",errormsg);
    exit(111); /* Exit code 111: Try again later */
    }

/* Harderror: Exit with an error code NOT allowing the MTA to try again */
/* As a general rule, use softerror when manipulations on non-user-controlled
   data go haywire, and use harderror when manipulations of user-controlled
   data go haywire */
int harderror(char *errormsg) {
    fprintf(stderr,"Fatal error: %s\n",errormsg);
    exit(100); /* Exit code 100: Return to sender */
    }

int main(int argc, char **argv) {
    js_string *arg, *tstr, *tstr2, *kiwirc;
    char badstr[256];
    char *envget; 
    int shift = 0;
    int tnum = 0;
    int offset = 0;

    /* Initialize js_string objects */
    if((arg = js_create(256,1)) == 0) 
        softerror("Unable to allocate arg.  Try later.");
    js_set_encode(arg,JS_US_ASCII);
    if((tstr = js_create(256,1)) == 0) 
        softerror("Unable to allocate tstr.  Try later.");
    js_set_encode(tstr,JS_US_ASCII);
    if((tstr2 = js_create(256,1)) == 0) 
        softerror("Unable to allocate tstr2.  Try later.");
    js_set_encode(tstr2,JS_US_ASCII);
    if((kiwirc = js_create(256,1)) == 0) 
        softerror("Unable to allocate kiwirc.  Try later.");
    js_set_encode(kiwirc,JS_US_ASCII);
    
    /* Regardless of anything else, we do need to read the user's .kiwirc
       file */
    if(find_kiwirc(kiwirc) == JS_ERROR)
        softerror("Error trying to locate .kiwirc file.");
    /* To do: Make sure the .kiwirc file has sane perms (600 or 700) */
    if(read_kiwirc(kiwirc,tstr,&tnum) == JS_ERROR)
        softerror("Error in reading .kiwirc file");
    if(tnum != 0) { /* If there was a syntax error in the .kiwirc file */
        if(tnum < -9000 || tnum > 9000)
	    softerror("Error number is out of bounds reading .kiwirc");
        sprintf(badstr,"%d",tnum); /* I really need a js_itoa function */
        js_qstr2js(tstr2,"Syntax error in .kiwirc file: ");
        js_append(tstr,tstr2); /* Add the error string */
        js_qstr2js(tstr,", line ");
        js_append(tstr,tstr2);
        js_qstr2js(tstr,badstr); /* Add the line number the error was on */
        js_append(tstr,tstr2);
        js_js2str(tstr2,badstr,100); /* Make it a NULL-terminated string */
        softerror(badstr);
        } 

    /* If argc is less than one (no arguments except argv[0]) we're haywire */
    if(argc < 1) 
        harderror("Argc is less than one!");
       
    /* Look at argv[0] and act accordingly */
    if(argv[0] != 0) {
        if(js_qstr2js(arg,argv[0]) == JS_ERROR) 
            harderror("Can not parse argv[0].");
        }
    else 
        harderror("argv[0] is a NULL pointer!");

    /* Now then, let's see how we should run Kiwi */

    /* Ignore anything leading up to a / */
    offset = 0;
    do {
        if(js_qstr2js(tstr,"/") == JS_ERROR)
            softerror("Can not give tstr a value, strange.");

        tnum = js_match_offset(tstr,arg,offset);
        if(tnum == JS_ERROR)
            harderror("Unable to run match on argv[0].");
    
        if(tnum != -2)
            offset = tnum + 1;

        } while(tnum != -2);

    /* See if the first letter of the command is an "i" (infilter) */

    if(js_qstr2js(tstr,"i") == JS_ERROR)
        softerror("Cannot give tstr a value!");

    tnum = js_match_offset(tstr,arg,offset);
    if(tnum == JS_ERROR)
        harderror("Unable to run match on argv[0]");
    if(tnum == offset) /* Mail filter for incoming mail */ {
	infilter(argc,argv);
        exit(0);
	}

    /* See if the first letter of the command is a "c" */

    if(js_qstr2js(tstr,"c") == JS_ERROR) 
        softerror("Can not give tstr a value, strange.");

    tnum = js_match_offset(tstr,arg,offset);

    if(tnum == JS_ERROR)
        harderror("Unable to run match on argv[0].");
    if(tnum == offset) /* Command-line cookie generation */ {
        envget = (char *)getenv("REMOTE_ADDR");
        if(argc == 1) /* No arguments (execpt argv[0]) */
            gencookie(envget); /* This will be NULL if REMOTE_ADDR not set */	
        else if(argc == 2) { /* One argument */
            gencookie(argv[1]);
	    printf("\n");
	    }
	exit(0);
        } 

    /* See if the first letter of the command is a "d" */

    if(js_qstr2js(tstr,"d") == JS_ERROR)
        softerror("Can not give tstr a value");

    tnum = js_match_offset(tstr,arg,offset);
    if(tnum == JS_ERROR)
        harderror("Unable to run match on argv[0]");
    if(tnum == offset) /* Decode cookie */ {
        if(argc == 1)
	    decode(0);
        else if(argc == 2) 
	    decode(argv[1]);
        exit(0);
	}

    /* See if the first letter of the command is a "w" (wrapper) */

    if(js_qstr2js(tstr,"w") == JS_ERROR)
        softerror("Can not give tstr a value!");

    tnum = js_match_offset(tstr,arg,offset);
    if(tnum == JS_ERROR)
        harderror("Unable to run match on argv[0]");
    if(tnum == offset) /* Sendmail wrapper for outgoing mail */ {
	wrapper(argc,argv);
        exit(0);
	}

    harderror("Invalid argv[0] value.");

    }

void decode(char *crypto) {
    js_string *cookie, *key, *format, *tstr;
    u_int32_t data;
    int counter, tnum;

    if((format = js_create(256,1)) == 0)
        softerror("Can not allocate temp string in gencookie");
    js_set_encode(format,JS_US_ASCII);
    if((tstr = js_create(256,1)) == 0)
        softerror("Can not allocate temp string in gencookie");
    js_set_encode(tstr,JS_US_ASCII);
    if((cookie = js_create(256,1)) == 0)
        softerror("Can not allocate temp string in gencookie");
    js_set_encode(cookie,JS_US_ASCII);
    if((key = js_create(256,1)) == 0)
        softerror("Can not allocate key string in gencookie");
    js_set_encode(key,JS_US_ASCII);

    /* Get and set up the encryption key */
    js_qstr2js(tstr,"kiwi_key");
    read_kvar(tstr,key);
    kjs_key_setup(key);

    if(crypto == 0) { /* Get cookie to encrypt from user input */
        js_getline_stdin(cookie);
        if(js_length(cookie) != COOKIE_SIZE + 1) /* We have a newline! */
            harderror("Encrypted cookie must be seven characters long");
        }
    else { /* Get it from crypto argument */
        js_qstr2js(cookie,crypto);
        if(js_length(cookie) != COOKIE_SIZE)
            harderror("Encrypted cookie must be seven characters long");
        }

    data = kjs_parse_cookie(cookie);
    data = kjs_grokdata(data,format);
    printf("%s","Data format: ");
    js_show_stdout(format);
    printf("\n");
   
    /* If the message is a timestamp, tell them what the time stamp is in
       both UNIX and English time */
    js_qstr2js(tstr,"TIMEOUT");
    tnum = js_fgrep(tstr,format);
    if(tnum == JS_ERROR)
        softerror("Strange problem determining format");
    if(tnum != -2) {
        /* Cut and paste from Kiwi version 1 */
        printf("Message data: %d, or %s",data,ctime((time_t *)&data));
	return;
        }
   
    /* If the message is an 16-IP range, tell them what IP range it is */
    js_qstr2js(tstr,"IP BLOCK");
    tnum = js_fgrep(tstr,format);
    if(tnum == JS_ERROR)
        softerror("Strange problem determining format.");
    if(tnum == 0) {
        /* Cut and paste from Kiwi version 1 */
        u_int32_t ip;
	u_int8_t ipo[4];
	ip = data;
	for(tnum=3;tnum>=0;tnum--) {
	    ipo[tnum] = ip & 0xff;
	    ip >>= 8;
	    }
        printf("Message: %d.%d.%d.%d-%d\n",ipo[0],ipo[1],ipo[2],
	       ipo[3],ipo[3]+16);
        return;
	}

    /* If the message is a five-letter embedded message (for mailing lists and 
       other permanent email addresses), tell them what the five letters are */
    js_qstr2js(tstr,"FIVE LETTERS");
    tnum = js_fgrep(tstr,format);
    if(tnum == JS_ERROR)
        softerror("A Strange problem determining format");
    if(tnum == 0) {
        /* Cut and paste from Kiwi version 1 */
        char mesg[6];
	for(tnum=4;tnum>=0;tnum--) {
	    mesg[tnum]=us_ascii_map(data & 31);
	    data >>= 5;
	    }
        mesg[5]=0;
	printf("Message: %s\n",mesg);
	}

    }
 
/* Generate a Kiwi cookies, and output them to the standard output.
   input: the ip/5-letter code we wish to encrypt (NULL if timestamp crypted)
   output: none (outputs information to stdout)
   global vars used: But only in kead_kvar that this calls
*/
void gencookie(char *tocrypt) {
    /* Set up the key for encrypting data */
    js_string *tstr, *key;
    u_int32_t data;
    time_t time_s;

    if((tstr = js_create(256,1)) == 0)
        softerror("Can not allocate temp string in gencookie");
    js_set_encode(tstr,JS_US_ASCII);
    if((key = js_create(256,1)) == 0)
        softerror("Can not allocate key string in gencookie");
    js_set_encode(key,JS_US_ASCII);

    /* Get and set up the encryption key */
    js_qstr2js(tstr,"kiwi_key");
    read_kvar(tstr,key);
    kjs_key_setup(key);

    if(tocrypt == 0) { /* Timestamp encryption */ 
        /* Get the time stamp and make cookies for short, mid, and long 
           timeout based on it */
        time_s = time(0);

        /* Short timeout */
        data = 0x80000000;
        data |= (time_s & 0xFFFFFFFF) >> 4; 
        kjs_make_cookie(data,tstr);
        printf("%s","Cookie with short timeout: ");
        js_show_stdout(tstr);
        printf("\n");

        /* Mid timeout */
        data = 0x90000000;
        data |= (time_s & 0xFFFFFFFF) >> 4; 
        kjs_make_cookie(data,tstr);
        printf("%s","Cookie with medium timeout (Usenet cookie): ");
        js_show_stdout(tstr);
        printf("\n");
 
        /* Long timeout */
        data = 0xA0000000;
        data |= (time_s & 0xFFFFFFFF) >> 4; 
        kjs_make_cookie(data,tstr);
        printf("%s","Cookie with long timeout: ");
        js_show_stdout(tstr);
        printf("\n");

        return;
        }
    else { /* Argument was given to tocrypt */

        if(js_qstr2js(tstr,tocrypt) == JS_ERROR)
	    harderror("Error converting REMOTE_ADDR/argv[1]");

        if(js_length(tstr) > 6) /* IP to crypt */ {
            int place = 0; /* Place to look for an octet */
            u_int8_t ipo[4];
            int counter;
	    js_string *dotq;
	    
	    if((dotq = js_create(256,1)) == 0)
	        softerror("Can not allocate memory for dotq");
           
	    js_set_encode(dotq,JS_US_ASCII);
	    js_qstr2js(dotq,".");

            for(counter = 0;counter < 4;counter++) {
                if(place == -2) 
		    harderror("IP must be in dotted octet form like 10.1.1.1");
                else {
                    if(place != 0)
                        place++;
	            ipo[counter] = js_atoi(tstr,place);
                    }
		place = js_match_offset(dotq,tstr,place);
		if(place == JS_ERROR)
		    harderror("Problem matching on IP string");
                }
            
	    data = (ipo[0] << 24) | (ipo[1] << 16) | (ipo[2] << 8) | ipo[3];
	    data >>= 4;
	    data |= 0xC0000000;
	    kjs_make_cookie(data,tstr);
	    js_show_stdout(tstr); 
	    return;
	    }
        else { /* 5-letter message to crypt */
	    int val,counter;
            data = 0;

            if(js_qstr2js(tstr,tocrypt) == JS_ERROR)
	        harderror("Unable to js-ize 5-letter message!");

            if(js_length(tstr) != 5)
	        harderror("Message must be exactly 5 letters long.");

            for(counter=0;counter<5;counter++) {
	        data <<= 5;

		val = js_val(tstr,counter);
                if(val == JS_ERROR)
                    harderror("Problem getting value of message char");
                
		val = us_ascii_unmap(val);
                if(val == JS_ERROR || val < 0 || val >= 26)
                    harderror("Please use only lower case letters in message");

                data |= val & 31;
                }

	    data |= 0x7C000000;
            kjs_make_cookie(data,tstr);
	    js_show_stdout(tstr);
	    return;
            }
        } 

    /* Something wrong if we get here */
    harderror("Invalid data given to clicrypt!");
    }

/* Sendmail wrapper for outgoing mail */
void wrapper(int argc, char **argv) {
    /* We need to know three things from .kiwirc: 
    
    1. What MTA we are running? (Qmail or Sendmail) 

    2. Our Kiwi key.

    3. The location of sendmail

    */

    js_string *qmailp, *key, *sendmail, *tstr, *cookie;

    u_int32_t data;

    int stream[2];

    /* Initialize the strings */
    if((qmailp = js_create(256,1)) == 0)
        softerror("Can not allocate qmailp string in wrapper");
    js_set_encode(qmailp,JS_US_ASCII);
    if((key = js_create(256,1)) == 0)
        softerror("Can not allocate key string in wrapper");
    js_set_encode(key,JS_US_ASCII);
    if((sendmail = js_create(256,1)) == 0)
        softerror("Can not allocate sendmail path string in wrapper");
    js_set_encode(sendmail,JS_US_ASCII);
    if((tstr = js_create(256,1)) == 0)
        softerror("Can not allocate tstr string in wrapper");
    js_set_encode(tstr,JS_US_ASCII);
    if((cookie = js_create(256,1)) == 0)
        softerror("Can not allocate cookie string in wrapper");
    js_set_encode(cookie,JS_US_ASCII);

    /* Obtain the values from the read .kiwirc file */
    js_qstr2js(tstr,"kiwi_qmail");
    read_kvar(tstr,qmailp);
    js_qstr2js(tstr,"kiwi_key");
    read_kvar(tstr,key);
    js_qstr2js(tstr,"kiwi_sendmail_path");
    read_kvar(tstr,sendmail);

    /* Generate the cookie to add to the email */
    data = 0xA0000000;
    data |= (time(0) & 0xFFFFFFFF) >> 4; 
    kjs_key_setup(key);
    if(kjs_make_cookie(data,cookie) == JS_ERROR)
        softerror("Problem setting up encrypted timestamp cookie");

    /* Set up an environmental variable to pass to the child */
    setenv("KIWI_STATUS","OUTGOING",1);

    /* Lots of cut and paste from Kiwi1 here, adapted to Kiwi2 */
    /* Set up things to pipe stdout to the sendmail child process */
    if(pipe(stream) != 0)
        softerror("Unable to initialize pipe");
   
    if(fork()) {
        /* Parent */
	char sendmail_cmd[256];
	close(stream[1]); /* close the output stream */
	dup2(stream[0],0); /* duplicate the standard input */
        if(js_js2str(sendmail,sendmail_cmd,200) == JS_ERROR)
	    harderror("Can not convert sendmail command to null-t string");
        execv(sendmail_cmd,argv);
	/* If we get here, there is a problem */
	harderror("Can not start sendmail!");
	}
    else { 
        /* Child */

	js_string *line, *tstr2, *lower;
	int inheader = 1;
        int addrline = 0;
	int showline, mungeline, match;

        if((line = js_create(256,1)) == 0)
            softerror("Can not allocate line string in wrapper");
        js_set_encode(line,JS_US_ASCII);
        if((tstr2 = js_create(256,1)) == 0)
            softerror("Can not allocate tstr2 string in wrapper");
        js_set_encode(tstr2,JS_US_ASCII);
        if((lower = js_create(256,1)) == 0)
            softerror("Can not allocate lower string in wrapper");
        js_set_encode(lower,JS_US_ASCII);
	
	close(stream[0]); /* Close the input stream */
	dup2(stream[1],1); /* duplicate the standard output to the pipe.
			      In other words, the standard output now goes
			      in to the pipe (sendmail) instead of to the
			      tty of the process */
        mungeline = showline = 1;
        do {
            
	    if(js_getline_stdin(line) == JS_ERROR)
	        harderror("Problem reading line from stdin");

            if(feof(stdin))
                break;

	    js_newline_chars(tstr);
	    js_space_chars(tstr2);
	    js_append(tstr2,tstr);
	    if(js_notmatch(tstr,line) == -2) /* If this is a blank line */
	        inheader = 0;

            js_qstr2js(tstr,AN); /* Alphanumeric characters */
            if(js_match(tstr,line) == 0) 
                mungeline = showline = 1;

	    /* Make lower a lower-case version of line */
            js_copy(line,lower);
	    js_tolower(lower);

	    /* Various lines in the heaqder we do not munge */
	    js_qstr2js(tstr,"message-id:");
	    if(js_fgrep(tstr,lower) == 0)
		mungeline = 0;
	    js_qstr2js(tstr,"to:");
	    if(js_fgrep(tstr,lower) == 0)
		mungeline = 0;
	    js_qstr2js(tstr,"cc:");
	    if(js_fgrep(tstr,lower) == 0)
		mungeline = 0;
	    js_qstr2js(tstr,"bcc:");
	    if(js_fgrep(tstr,lower) == 0)
		mungeline = 0;
	    js_qstr2js(tstr,"content-");
	    if(js_fgrep(tstr,lower) == 0)
		mungeline = 0;
	    js_qstr2js(tstr,"in-reply-to:");
	    if(js_fgrep(tstr,lower) == 0)
		mungeline = 0;
	    js_qstr2js(tstr,"subject:");
	    if(js_fgrep(tstr,lower) == 0)
		mungeline = 0;
            /* We do not show an X-Sender: line */	
            js_qstr2js(tstr,"x-sender:"); 
            if(js_fgrep(tstr,lower) == 0)
		showline = 0;

            if(mungeline && inheader) {
	    	js_qstr2js(tstr,"@");
	        match = js_match(tstr,line);
		if(match == JS_ERROR)
		    harderror("Problem running match against line");
	
		if(match != -2) {
		    /* Inset actual Kiwi cookie */
		    if(js_insert(cookie,line,match) == JS_ERROR)
		        harderror("Problem inserting Kiwi cookie");

		    /* Then inset -/+ (Done this way because insert inserts
                       _after_ the string place in question) */
		    js_qstr2js(tstr,"TRUE");
		    if(js_fgrep(tstr,qmailp) == 0) {
		        js_qstr2js(tstr,"-");
			if(js_insert(tstr,line,match) == JS_ERROR)
			    harderror("Problem inserting qmail dash");
                        }
                    else {
		        js_qstr2js(tstr,"+");
			if(js_insert(tstr,line,match) == JS_ERROR)
			    harderror("Problem inserting sendmail dash");
                        }
		    }
		}  

	    if(showline) 
	        js_show_stdout(line);

	    } while(!feof(stdin));
	}
    /* Both parent and child */
    exit(0); 
    }

/* Getaddress: Get an email address from a mail header line
   Input: Pointer to line, pointer to js_string to place address in
   Output: JS_ERROR on error, JS_SUCCESS on success 
   (address will be null if address not found) */
int getaddress(js_string *line, js_string *address) {

    js_string *exp, *tstr;
    int begin, end, place;

    /* Sanitry checks */
    if(kiwi_goodjs(line) == JS_ERROR)
        return JS_ERROR;
    if(kiwi_goodjs(address) == JS_ERROR)
        return JS_ERROR;

    /* Blank out address */
    js_qstr2js(address,"");

    /* Allocate the expression we match against */
    if((exp = js_create(256,1)) == 0)
        return JS_ERROR;
    js_set_encode(exp,JS_US_ASCII);
    /* And a temporary string */
    if((tstr = js_create(256,1)) == 0)
        return JS_ERROR;
    js_set_encode(tstr,JS_US_ASCII);

    /* Look for anything with an @ in it between <> */
    if(js_qstr2js(exp,"<") == JS_ERROR)
        return JS_ERROR;
    if((begin = js_fgrep(exp,line)) == JS_ERROR)
        return JS_ERROR;
    if(begin != -2) { /* < found */
        if(js_qstr2js(exp,">") == JS_ERROR)
	    return JS_ERROR;
        if((end = js_fgrep_offset(exp,line,begin)) == JS_ERROR)
	    return JS_ERROR;
        if(end != -2) { /* > found after < */
	    if((js_substr(line,address,begin + 1,end - begin - 1)) == JS_ERROR)
	        return JS_ERROR;
            if(js_qstr2js(exp,"@") == JS_ERROR)
	        return JS_ERROR;
            if((place = js_fgrep(exp,address)) == JS_ERROR)
	        return JS_ERROR;
            if(place != -2) /* @ in address */
	        return JS_SUCCESS;
            }
        }

    /* If <> search fails, look for bare word with an @ in it */
    if(js_qstr2js(exp,"@") == JS_ERROR)
        return JS_ERROR;
    if((place = js_fgrep(exp,line)) == JS_ERROR)
        return JS_ERROR;
    if(place == -2) /* No @ found */
        return JS_SUCCESS; /* Address string is 0-length string */

    /* Start looking for whitespace/newlines */
    js_space_chars(tstr);
    js_newline_chars(exp);
    js_append(tstr,exp);
    end = -1; /* Doesn't mean error--this makes code elegant */
    while(end < place) {
        begin = end;
	if((end = js_match_offset(exp,line,begin + 1)) == JS_ERROR)
	    return JS_ERROR;
        if(end == -2) { /* @ in word at end of string */ 
	    end = js_length(line) + 1;
	    break;
	    }
        }

    /* The word is between the last whitespace before the @ (or the
       beginning of the string) and the first whitespace after the @ 
       (or the end of the string) */
    if((js_substr(line,address,begin + 1,end - begin - 1)) == JS_ERROR)
        return JS_ERROR;
    return JS_SUCCESS;
    }

/* parseaddress: Parse the kiwi-cookie in an email address 
   input: email address to parse
   output: data value of cookie in address
*/
u_int32_t parseaddress(js_string *address) {
    /* We need to use the following kvars:

       kiwi_qmail: if we are using Qmail or sendmail

       kiwi_password: if we have a valid kiwi password */

    js_string *qmailp, *password, *tstr, *exp;

    int dash, at, is_qmail = 0, temp;

    u_int32_t data;

    /* Initialize the strings */
    if((qmailp = js_create(256,1)) == 0)
        softerror("Can not allocate qmailp string in wrapper");
    js_set_encode(qmailp,JS_US_ASCII);
    if((password = js_create(256,1)) == 0)
        softerror("Can not allocate key string in wrapper");
    js_set_encode(password,JS_US_ASCII);
    if((tstr = js_create(256,1)) == 0)
        softerror("Can not allocate key string in wrapper");
    js_set_encode(tstr,JS_US_ASCII);
    if((exp = js_create(256,1)) == 0)
        softerror("Can not allocate key string in wrapper");
    js_set_encode(exp,JS_US_ASCII);

    /* Obtain the values from the read .kiwirc file */
    js_qstr2js(tstr,"kiwi_qmail");
    read_kvar(tstr,qmailp);
    js_qstr2js(tstr,"kiwi_password");
    read_kvar(tstr,password);

    /* Find the "@" and the last dash before the "@" */
    js_qstr2js(exp,"@");
    at = js_fgrep(exp,address);
    if(at == JS_ERROR)
        return 0;
    if(at == -2)
        return 1; /* No cookie found -- invalid */
    /* See if we are running Qmail or Sendmail */
    js_qstr2js(exp,"true");
    js_copy(qmailp,tstr);
    js_tolower(tstr);
    if(js_fgrep(exp,tstr) == 0)
        is_qmail = 1;
    /* Now, look for the dash */
    if(is_qmail)
        js_qstr2js(exp,"-");
    else
        js_qstr2js(exp,"+");
    temp = 0;
    dash = -2;
    while(temp < at && temp != -2) {
        dash = temp;
	if(temp < js_length(address) - 2)
            temp = js_fgrep_offset(exp,address,temp + 1);
        else
	    temp = -2;
        if(temp == JS_ERROR)
	    return 0;
        }
    if(dash == -2) /* No dash found before at */
        return 1; /* No cookie -- invalid */
    if(at - dash < 2) /* invalid values */
        return 1;
    /* Make tstr the cookie we want to grok */
    if(js_substr(address,tstr,dash + 1,at - dash - 1) == JS_ERROR)
        return 0;
    /* Password check. Todo: Make case-insensitive */
    if(js_issame(tstr,password) == 1)
        return 2;
    data = kjs_parse_cookie(tstr);
    if(data == 0 || data == 1 || data == 2)
        return 1;
    return data; 
    }

/* Mail filter for incoming mail */
void infilter(int argc, char **argv) {
    js_string *key, *tstr;
    js_string *foo, *bar; /* DEBUG */
    u_int32_t baz; /* DEBUG */

    if((key = js_create(256,1)) == 0)
        softerror("Can not allocate qmailp string in wrapper");
    js_set_encode(key,JS_US_ASCII);
    if((tstr = js_create(256,1)) == 0)
        softerror("Can not allocate qmailp string in wrapper");
    js_set_encode(tstr,JS_US_ASCII);

    js_qstr2js(tstr,"kiwi_key");
    read_kvar(tstr,key); 
    kjs_key_setup(key);

    /* DEBUG CODE */
    foo = js_create(256,1);
    js_set_encode(foo,JS_US_ASCII);  
    bar = js_create(256,1);
    js_set_encode(bar,JS_US_ASCII);  
    while(!feof(stdin)) {
        js_getline_stdin(foo);
        baz = parseaddress(foo);
	printf("%d\n",baz);
	}
    exit(0); 
    }


